--
-- video.vhd - Top Level file for the Video Chip project
--
-- Provide replacement for TMS9929A, which outputs to VGA.
-- Targetting GODIL boards with Spartan 3E model XC3S500E.
--

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use IEEE.NUMERIC_STD.ALL;

entity video is
  port
    (
    -- inputs only
    m49  : in    std_logic; -- 49.152MHz clock
    sw1  : in    std_logic; -- press to disable ROM 4
    sw2  : in    std_logic; -- normally jumpered so as to reset FPGA
    sout : in    std_logic; -- red LED 4, not usable
    rts  : in    std_logic; -- red LED 5, not usable
    -- I/O's for UART and SPI flash
    vs2  : inout std_logic; -- green LED 8
    cts  : inout std_logic; -- red LED 6, not usable, and yet its inout
    cso  : inout std_logic; -- red LED 9, SPI configuration memory chip select
    sin  : inout std_logic; -- red LED 3
    -- I/O's for test connector
    tvs1,tvs0,tmosi,tdin,tcclk,tm1,thsw : inout std_logic;
    -- Configuration
    LHS_IN       : in    std_logic;
    RHS_IN       : in    std_logic;
    -- CPU interface
    RESET_n      : in    std_logic;
    CSR_n        : in    std_logic; -- CPU-VDP read strobe
    CSW_n        : in    std_logic; -- CPU-VDP write strobe
    MODE         : in    std_logic; -- data verses control
    CD           : inout std_logic_vector(7 downto 0); -- CPU data bus
    INT_n        : out   std_logic; -- interrupt
    -- crystal inputs, not used
    XTAL1        : in    std_logic;
    XTAL2        : in    std_logic;
    -- VRAM interface, not used
    READ         : out   std_logic; -- VRAM write strobe
    AD           : out   std_logic_vector(7 downto 0); -- address to VRAM
    RAS_n        : out   std_logic;
    CAS_n        : out   std_logic;
    RD           : in    std_logic_vector(7 downto 0); -- VRAM read data bus
    -- video output, not used
    Y            : out   std_logic;
    R_Y          : out   std_logic;
    B_Y          : out   std_logic;
    GROMCLK      : out   std_logic;
    -- VGA interface
    VGA_R        : out   std_logic_vector(3 downto 0);
    VGA_G        : out   std_logic_vector(3 downto 0);
    VGA_B        : out   std_logic_vector(3 downto 0);
    VGA_HS       : out   std_logic;
    VGA_VS       : out   std_logic;
    -- power
    VCC          : in    std_logic;
    VSS          : in    std_logic
    );
end video;

architecture behavior of video is

  component clock_divider
    port
      (
      clk_50mhz : in  std_logic;
      clk_25mhz : out std_logic
      );
  end component;

  component vdp_vram
    port
      (
      clk_host  : in  std_logic;
      d_host    : in  std_logic_vector(7 downto 0);
      addr_host : in  std_logic_vector(13 downto 0);
      we_host   : in  std_logic;
      q_host    : out std_logic_vector(7 downto 0);
      clk_vdp   : in  std_logic;
      addr_vdp  : in  std_logic_vector(13 downto 0);
      q_vdp     : out std_logic_vector(7 downto 0)
      );
  end component;

  component vdp
    port
      (
      clk_25mhz     : in  std_logic;
      vram_addr     : out std_logic_vector(13 downto 0);
      vram_q        : in  std_logic_vector(7 downto 0);
      clk_cpu       : in  std_logic;
      reg_re        : in  std_logic;
      reg_we        : in  std_logic;
      reg_number    : in  std_logic_vector(2 downto 0);
      reg_value     : in  std_logic_vector(7 downto 0);
      status        : out std_logic_vector(7 downto 0);
      interrupt     : out std_logic;
      hw_pal        : in  std_logic;
      frame_rate    : in  std_logic;
      debug_patnam  : in  std_logic;
      debug_patgen  : in  std_logic;
      debug_patcol  : in  std_logic;
      debug_sprgen  : in  std_logic;
      vga_r         : out std_logic_vector(3 downto 0);
      vga_g         : out std_logic_vector(3 downto 0);
      vga_b         : out std_logic_vector(3 downto 0);
      vga_hsync     : out std_logic;
      vga_vsync     : out std_logic
      );
  end component;

  -- clock
  signal clk_25mhz : std_logic := '0';

  -- VRAM related
  signal host_vram_addr     : std_logic_vector(13 downto 0) := "00000000000000";
  signal host_vram_addr_inc : std_logic := '0';
  signal host_vram_d        : std_logic_vector(7 downto 0);
  signal host_vram_we       : std_logic := '0';
  signal host_vram_q        : std_logic_vector(7 downto 0);
  signal vdp_vram_addr      : std_logic_vector(13 downto 0);
  signal vdp_vram_q         : std_logic_vector(7 downto 0);

  -- VDP chip related
  signal vdp_latched     : std_logic := '0';
  signal vdp_reg_re      : std_logic;
  signal vdp_reg_we      : std_logic := '0';
  signal vdp_reg_number  : std_logic_vector(2 downto 0);
  signal vdp_reg_value   : std_logic_vector(7 downto 0);
  signal vdp_status      : std_logic_vector(7 downto 0);
  signal vdp_interrupt   : std_logic;

  -- relating to the CPU / VDP interface
  signal CSR_n_prev      : std_logic_vector(3 downto 0) := "1111";
  signal CSW_n_prev      : std_logic_vector(3 downto 0) := "1111";
  signal MODE_latched    : std_logic;

  -- deglitching the RESET_n signal
  -- 4095 consequtive samples at 49MHz rate (ie: 83us)
  signal RESET_n_counter : std_logic_vector(11 downto 0) := "000000000000";
  signal RESET_n_clean   : std_logic := '1';

begin

  U_CLOCK : clock_divider
    port map
      (
      clk_50mhz => m49,       -- well, its kinda close to 50MHz
      clk_25mhz => clk_25mhz  -- and so this'll be close to 25MHz VGA dot clock
      );

  U_VRAM : vdp_vram
    port map (
      clk_host  => m49,
      d_host    => host_vram_d,
      addr_host => host_vram_addr,
      we_host   => host_vram_we,
      q_host    => host_vram_q,
      clk_vdp   => clk_25mhz,
      addr_vdp  => vdp_vram_addr,
      q_vdp     => vdp_vram_q
      );

  U_VDP : vdp
    port map
      (
      clk_25mhz     => clk_25mhz,
      vram_addr     => vdp_vram_addr,
      vram_q        => vdp_vram_q,
      clk_cpu       => m49,
      reg_re        => vdp_reg_re,
      reg_we        => vdp_reg_we,
      reg_number    => vdp_reg_number,
      reg_value     => vdp_reg_value,
      status        => vdp_status,
      interrupt     => vdp_interrupt,
      hw_pal        => LHS_IN,
      frame_rate    => RHS_IN,
      debug_patnam  => sw1,
      debug_patgen  => '0',
      debug_patcol  => '0',
      debug_sprgen  => '0',
      vga_r(3)      => VGA_R(3),
      vga_r(2)      => VGA_R(2),
      vga_r(1)      => open,
      vga_r(0)      => open,
      vga_g(3)      => VGA_G(3),
      vga_g(2)      => VGA_G(2),
      vga_g(1)      => open,
      vga_g(0)      => open,
      vga_b(3)      => VGA_B(3),
      vga_b(2)      => VGA_B(2),
      vga_b(1)      => open,
      vga_b(0)      => open,
      vga_hsync     => VGA_HS,
      vga_vsync     => VGA_VS
      );

  -- clean up the RESET_n signal
  process ( m49, RESET_n )
  begin
    if rising_edge(m49) then
      if RESET_n = '0' and RESET_n_counter = "111111111111" then
        RESET_n_clean <= '0';
      elsif RESET_n = '0' then
        RESET_n_counter <= RESET_n_counter+1;
      else
        RESET_n_counter <= "000000000000";
        RESET_n_clean <= '1';
      end if;
    end if;
  end process;    

  -- IO writes, and side effects of reads
  process ( m49, RESET_n_clean )
  begin
    if rising_edge(m49) then
      if RESET_n_clean = '0' then
        host_vram_addr     <= "00000000000000";
        host_vram_we       <= '0';
        host_vram_addr_inc <= '0';
        vdp_latched        <= '0';
        vdp_reg_we         <= '0';
      else
        host_vram_we       <= '0';
        if host_vram_addr_inc = '1' then
          host_vram_addr <= host_vram_addr+1;
        end if;
        host_vram_addr_inc <= '0';
        vdp_latched        <= vdp_latched;
        vdp_reg_we         <= '0';

        if CSW_n_prev = "1110" and CSW_n = '0' then
          -- MODE is required to be held for 30ns after CSW_n goes '0'
          -- due to our deglitching logic, we'll notice within 40ns
          MODE_latched <= MODE;
        elsif CSW_n_prev = "0001" and CSW_n = '1' then
          -- CD is required to be held for 30ns after CSW_n goes '1'
          -- due to our deglitching logic, we'll notice within 40ns
          case MODE_latched is
            when '0' =>
              -- VDP data
              -- present data to VRAM
              -- on next clock it'll be latched
              host_vram_d        <= CD;
              host_vram_we       <= '1';
              host_vram_addr_inc <= '1';
              vdp_latched        <= '0';
            when '1' =>
              -- VDP control
              if vdp_latched = '0' then
                host_vram_addr(7 downto 0) <= CD;
                vdp_latched <= '1';
              else
                case CD(7 downto 6) is
                  when "00" | "01" =>
                    -- set up for reading or writing
                    host_vram_addr(13 downto 8) <= CD(5 downto 0);
                  when "10" =>
                    -- write VDP register
                    vdp_reg_number <= CD(2 downto 0);
                    vdp_reg_value  <= host_vram_addr(7 downto 0);
                    vdp_reg_we     <= '1';
                  when others =>
                end case;
                vdp_latched <= '0';
              end if;
            when others =>
              -- nothing
          end case;
        elsif CSR_n_prev = "0001" and CSR_n = '1' then
          -- Host has used the data
          case MODE is
            when '0' =>
              -- VDP data
              host_vram_addr_inc <= '1';
              vdp_latched        <= '0';
            when '1' =>
              -- VDP status
              vdp_latched <= '0';
            when others =>
              -- nothing
          end case;
        end if;
      end if;
      CSR_n_prev <= CSR_n_prev(2 downto 0) & CSR_n;
      CSW_n_prev <= CSW_n_prev(2 downto 0) & CSW_n;
    end if; -- rising_edge(m49)
  end process;

  -- Ensure VDP knows of CPU intent to read status
  vdp_reg_re <= '1' when ( CSR_n = '0' and MODE = '1' ) else '0';

  -- Only drive CPU data bus when we've got some data,
  -- and we're allowed to drive it

  CD <= host_vram_q
    when ( CSR_n = '0' and MODE = '0' )

  else vdp_status
    when ( CSR_n = '0' and MODE = '1' )

  else "ZZZZZZZZ";

  INT_n <= not vdp_interrupt;

  -- values for those things we don't care about

  READ    <= '1';
  AD      <= "00000000";
  RAS_n   <= '1';
  CAS_n   <= '1';
  Y       <= '0';
  R_Y     <= '0';
  B_Y     <= '0';
  GROMCLK <= '0';

  -- try to set up LEDs

  vs2   <= '1'; -- green LED
  cts   <= '1'; -- red LED 6 top middle left and handshake to non existent USB
  cso   <= '1'; -- red LED 9 and SPI configuration memory chip select
  sin   <= '1'; -- red LED 3 top left

  -- try to do sensible things with the other pins

  tvs1  <= '1';
  tvs0  <= '1';
  tmosi <= '1';
  tdin  <= 'Z'; -- don't drive it, its an input
  tcclk <= '1'; -- not talking to SPI Flash
  tm1   <= '1'; -- not configuring
  thsw  <= '1'; -- not configuring

end behavior;
